const url = require("url");
const config = require("../config");
const Signature = require("./signature");

const channelPattern = new RegExp(
	"^([a-f0-9]{32})(?:\\:([a-f0-9]{32}))?(?:\\.([a-f0-9]{32,128}))?$"
);

class Channel
{
	/**
	 *
	 * @param {Buffer} privateId
	 * @param {?Buffer} publicId
	 */
	constructor(privateId, publicId)
	{
		/**
		 *
		 * @type {Buffer}
		 */
		this.privateId = privateId;

		/**
		 *
		 * @type {?Buffer}
		 */
		this.publicId = Buffer.isBuffer(publicId) ? publicId : null;

		/**
		 *
		 * @type {?String}
		 */
		this.hexPrivateId = null;
		this.hexPublicId = null;
	}

	getPrivateId()
	{
		return this.privateId;
	}

	getHexPrivateId()
	{
		if (this.hexPrivateId === null)
		{
			this.hexPrivateId = this.privateId.toString("hex");
		}

		return this.hexPrivateId;
	}

	getPublicId()
	{
		return this.publicId;
	}

	getHexPublicId()
	{
		if (this.hexPublicId === null)
		{
			this.hexPublicId = this.publicId.toString("hex");
		}

		return this.hexPublicId;
	}

	/**
	 *
	 * @param request
	 * @param {Boolean} [skipSign=false]
	 * @returns Channel[]
	 */
	static createFromRequest(request, skipSign)
	{
		const channels = [];
		const uri = url.parse(request.url, true);
		const channelIds = uri.query.CHANNEL_ID;
		const clientId =
			config.trustClients === false &&
			typeof(uri.query.clientId) === "string" &&
			uri.query.clientId.match(/^[a-f0-9]{32}$/)
				? uri.query.clientId
				: null
		;

		if (!channelIds || channelIds.length < 1)
		{
			return [];
		}

		if (skipSign !== true && clientId === null && config.trustClients === false)
		{
			return [];
		}

		const channelParts = channelIds.split("/");
		for (let i = 0; i < channelParts.length; i++)
		{
			let result = Channel.parse(channelParts[i], skipSign, clientId);
			if (!result)
			{
				return [];
			}

			const [privateId, publicId] = result;

			channels.push(
				new Channel(
					Buffer.from(privateId, "hex"),
					publicId ? Buffer.from(publicId, "hex") : null
				)
			);
		}

		return channels;
	}

	/**
	 *
	 * @param channel
	 * @param skipSign
	 * @param clientId
	 * @returns Boolean|Array
	 */
	static parse(channel, skipSign, clientId = null)
	{
		const match = typeof(channel) === "string" && channel.match(channelPattern);
		if (!match)
		{
			return false;
		}

		const [, privateChannelId, publicChannelId, signature] = match;

		if (publicChannelId)
		{
			if (!signature || !Signature.isValid(privateChannelId + ":" + publicChannelId, signature, clientId))
			{
				return false;
			}
		}
		else if (skipSign !== true && (!signature || !Signature.isValid(privateChannelId, signature, clientId)))
		{
			return false;
		}

		return [privateChannelId, publicChannelId];
	}

	static isValid(id)
	{
		return Buffer.isBuffer(id) && id.length === 16;
	}
}

module.exports = Channel;