const Sqlite = require("./storages/sqlite");
const superAgent = require("superagent");
const crypto = require("crypto");
const config = require("../config");
const logger = require("./debug");
const { RegisterRequest, Notification, License } = require("./models");
let application = null;

const cache = new Map();
const dateTimePattern = /^(20[0-9]{2})-([0-9]{2})-([0-9]{2})$/;


//Let's extend License class with static methods.

/**
 *
 * @param {string} clientId
 * @param useCache
 * @return null|License
 */
License.get = function(clientId, useCache = true)
{
	if (typeof (clientId) !== "string" || clientId.length < 32)
	{
		return null;
	}

	let license = useCache ? cache.get(clientId) : undefined;
	if (license === undefined)
	{
		try
		{
			const db = Sqlite.getDatabase();
			if (db !== null)
			{
				const statement = db.prepare("SELECT * FROM licenses WHERE clientId = ?");
				const row = statement.get(clientId);

				license = row ? License.create(row) : null;
				cache.set(clientId, license);
			}
			else
			{
				license = null;
			}
		}
		catch (error)
		{
			logger.systemError(error);
			license = null;
		}
	}

	return license;
};

/**
 *
 * @param {RegisterRequest} registerRequest
 * @param {function<Error, License>} callback
 */
License.register = function(registerRequest, callback)
{
	superAgent
		.post(config.licenseServer)
		.send(registerRequest.verificationQuery)
		.retry(2)
		.timeout({
			response: 5000,
			deadline: 10000
		})
		.then(response => {

			const body = response.body || {};
			const result = body.result && typeof (body.result) === "object" ? body.result : null;

			if (body.status === "ok" && result)
			{
				const license = License.create({
					clientId: result["LICENSE_KEY_HEAD"],
					active: result["ACTIVE"],
					dateTo: result["DATE_TO"],
					siteUrl: result["SITE_URL"],
					verificationQuery: registerRequest.verificationQuery
				});

				License.save(license, callback);
			}
			else
			{
				const error = body.status === "error" && body.text ? body.text : "Unknown Error";

				callback(new Error(error), null);
			}

		})
		.catch(error => {
			callback(error, null);
		})
	;
};

License.unregister = function(clientId, callback)
{
	if (typeof(clientId) !== "string" || !clientId.match(/[a-z0-9]{32}/))
	{
		callback(new Error("Wrong Client Id."), null);
		return;
	}

	License.refresh(clientId);
	const license = License.get(clientId, false);
	if (license === null)
	{
		callback(new Error("Client not found."), null);
		return;
	}

	const db = Sqlite.getDatabase();
	if (!db)
	{
		callback(new Error("Couldn't connect to the database."), null);
		return;
	}

	const statement = db.prepare(`DELETE FROM licenses WHERE clientId = @clientId`);

	try
	{
		statement.run({
			clientId: clientId,
		});

		if (application === null)
		{
			application = require("./application");
		}

		application.getAdapter().postIpcNotification(Notification.create({
			ipcLicenses: {
				licenses: [{
					license,
					action: "unregister"
				}]
			}
		}));

		callback(null, license);
	}
	catch (error)
	{
		callback(error, null);
	}
};

/**
 *
 * @param {License} license
 * @param {function<Error, License>} callback
 */
License.save = function(license, callback)
{
	const db = Sqlite.getDatabase();
	if (!db)
	{
		callback(new Error("Couldn't connect to the database."), null);
		return;
	}

	license = license && typeof (license) === "object" ? license : {};

	const clientId = typeof (license.clientId) === "string" ? license.clientId : "";
	if (!clientId.match(/[a-z0-9]{32}/))
	{
		callback(new Error("Wrong Client ID."), null);
		return;
	}

	//Test Client ID
	let securityKey = crypto.randomBytes(64).toString("hex");
	if (clientId === "fd818684484258a5c6f0442a070661d6")
	{
		if (
			config.security &&
			typeof (config.security.key) === "string" &&
			config.security.key.length >= 32 &&
			config.security.key.length <= 512
		)
		{
			securityKey = config.security.key;
		}
	}

	let dateTo = typeof (license.dateTo) === "string" ? license.dateTo : "";
	const result = dateTimePattern.exec(dateTo);
	if (result === null)
	{
		callback(new Error("Wrong expiry date."), null);
		return;
	}

	dateTo = (new Date(dateTo)).getTime();

	const siteUrl = typeof (license.siteUrl) === "string" ? license.siteUrl : "";
	if (siteUrl.length > 100)
	{
		callback(new Error("Site URL is too long (100)."), null);
		return;
	}

	const verificationQuery = typeof (license.verificationQuery) === "string" ? license.verificationQuery : "";
	if (verificationQuery.length < 64 || verificationQuery.length > 200)
	{
		callback(new Error("Wrong Verification Query."), null);
		return;
	}

	const statement = db.prepare(`
		INSERT INTO licenses (clientId, securityKey, dateTo, siteUrl, verificationQuery)
		VALUES (@clientId, @securityKey, @dateTo, @siteUrl, @verificationQuery)
		ON CONFLICT(clientId)
		DO UPDATE SET
			dateTo=excluded.dateTo,
			siteUrl=excluded.siteUrl,
			verificationQuery=excluded.verificationQuery
	`);

	try
	{
		statement.run({
			clientId: clientId,
			securityKey: securityKey,
			dateTo: dateTo,
			siteUrl: siteUrl,
			verificationQuery: verificationQuery
		});

		License.refresh(clientId);
		const license = License.get(clientId);

		if (application === null)
		{
			application = require("./application");
		}

		application.getAdapter().postIpcNotification(Notification.create({
			ipcLicenses: {
				licenses: [{
					license,
					action: "register"
				}]
			}
		}));

		callback(null, license);
	}
	catch (error)
	{
		callback(error, null);
	}
};

/**
 *
 * @param {string|License} license
 */
License.isValid = function(license)
{
	if (typeof (license) === "string")
	{
		license = License.get(license);
	}

	if (license && license.clientId)
	{
		if (!Number.isInteger(license.dateTo))
		{
			return false;
		}

		const delta = license.dateTo - Date.now();
		if (delta > 0)
		{
			return true;
		}
		else
		{
			const oneDay = 3600 * 24 * 1000;
			const twoWeeks = 3600 * 24 * 14 * 1000;

			if (Number.isInteger(license.lastCheck) && (Date.now() - license.lastCheck) > oneDay)
			{
				License.sync(license.clientId);
			}

			if (delta > -twoWeeks)
			{
				return true;
			}
		}
	}

	return false;
};

License.sync = function(clientId)
{
	const license = License.get(clientId, false);
	const oneDay = 3600 * 24 * 1000;

	if (license === null || (Date.now() - license.lastCheck) < oneDay)
	{
		return;
	}

	const db = Sqlite.getDatabase();
	if (!db)
	{
		return;
	}

	try
	{
		const statement = db.prepare(`UPDATE licenses SET lastCheck=@lastCheck WHERE clientId = @clientId`);
		statement.run({
			lastCheck: Date.now(),
			clientId: clientId
		});

		license.lastCheck = Date.now();
	}
	catch (error)
	{
		logger.systemError("License refresh failed.", error);
		return;
	}

	const registerRequest = RegisterRequest.create({
		verificationQuery: license.verificationQuery
	});

	License.register(registerRequest, (error, /*License*/license) => {
		if (error)
		{
			logger.systemError("License refresh failed.", error);
		}
	});
};

License.refresh = function(clientId)
{
	cache.delete(clientId);
};

module.exports = License;